@@ -16,6 +16,16 @@ from keras.callbacks import ModelCheckpoint
 
 
 class MultiGPUModelCheckpoint(ModelCheckpoint):
+
+    def __init__(self, *args, **kwargs):
+        super(MultiGPUModelCheckpoint,self).__init__(*args,**kwargs)
+        self.template_model = None
+        self.best = None
+        self.fine_tune = None
+        self.anchors - None
+        self.num_free_layers = None
+        self.version = None
+
     def set_template_model(self, template_model):
         self.template_model = template_model
 
@@ -29,6 +39,9 @@ class MultiGPUModelCheckpoint(ModelCheckpoint):
     def reset_best(self, best):
         self.best = best
 
+    def set_version(self, version):
+        self.version = version
+
     def on_epoch_end(self, epoch, logs=None):
         current = logs.get(self.monitor)
         if current is not None:
@@ -55,6 +68,9 @@ class MultiGPUModelCheckpoint(ModelCheckpoint):
                     if self.fine_tune:
                         f["num_free_layers"] = [self.num_free_layers]
 
+                    if self.version is not None:
+                        f["cryolo_version"] = [str(self.version).encode('utf8')]
+
             else:
                 if self.verbose > 0:
                     print(
