@@ -22,13 +22,11 @@ Prediction script crYOLO
 #
 from __future__ import print_function
 import multiprocessing
-
 import time
 import argparse
 import os
 import sys
 import json
-import traceback
 import numpy as np
 from lineenhancer import line_enhancer, maskstackcreator
 from . import CoordsIO
@@ -52,6 +50,7 @@ os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
 ARGPARSER = None
 filament_tracers = []
 
+
 def create_parser(parser, use_gooey=True):
     """
 
@@ -219,6 +218,20 @@ def create_parser(parser, use_gooey=True):
         help="On the fly filtering. Filtered micrographs will not be written to disk. Might be slower",
     )
 
+    optional_group.add_argument(
+        "--cleanup",
+        action="store_true",
+        default=False,
+        help="If true, it will delete the filtered images after training is done."
+    )
+
+    optional_group.add_argument(
+        "--skip",
+        action="store_true",
+        default=False,
+        help="If true, it will skip images that were already picked."
+    )
+
     filament_group = parser.add_argument_group(
         "Filament options",
         "These options are only relevant if you want to use  the filament mode",
@@ -329,11 +342,20 @@ def _main_():
 
 
 def main(args=None):
+    start_method = "fork"
     try:
-        multiprocessing.set_start_method("fork")
+        os_start_method = os.environ["CRYOLO_MP_START"]
+
+        if os_start_method in ["spawn", "fork"]:
+            start_method = os_start_method
+    except:
+        pass
+    try:
+        multiprocessing.set_start_method(start_method)
     except RuntimeError:
         pass
     import cryolo.utils as util
+
     util.check_for_updates()
 
     print("START METHOD:", multiprocessing.get_start_method())
@@ -404,7 +426,7 @@ def main(args=None):
     write_empty = args.write_empty
     min_size = args.minsize
     max_size = args.maxsize
-    num_cpus = int(multiprocessing.cpu_count()/2)
+    num_cpus = int(multiprocessing.cpu_count() / 2)
     if args.num_cpu != -1:
         num_cpus = args.num_cpu
 
@@ -413,15 +435,10 @@ def main(args=None):
             config = json.load(config_buffer)
         except json.JSONDecodeError:
             print(
-                "Your configuration file seems to be corruped. Please check if it is valid."
+                "Your configuration file seems to be corrupted. Please check if it is valid."
             )
-        if config["model"]["input_size"] % 32 > 0:
-            new_input_size = int(32 * round(float(config["model"]["input_size"]) / 32))
-            print(
-                "You input size is not a multiple of 32. Round it to the next multiple of 32:",
-                new_input_size,
-            )
-            config["model"]["input_size"] = new_input_size
+        config["model"]["input_size"] = config_tools.get_adjusted_input_size(config)
+
     # Setup log dir
     if "other" in config and "log_path" in config["other"]:
         log_path = config["other"]["log_path"]
@@ -431,10 +448,12 @@ def main(args=None):
         os.makedirs(log_path)
 
     # Write command to disk
-    import time
+
     timestr = time.strftime("%Y%m%d-%H%M%S")
-    utils.write_command(os.path.join(log_path,"cmdlogs/", "command_predict_" + timestr + ".txt"),
-                        "cryolo_predict.py " + ' '.join(sys.argv[1:]))
+    utils.write_command(
+        os.path.join(log_path, "cmdlogs/", "command_predict_" + timestr + ".txt"),
+        "cryolo_predict.py " + " ".join(sys.argv[1:]),
+    )
 
     filament_mode = args.filament
     if filament_mode:
@@ -463,6 +482,10 @@ def main(args=None):
         if not len(config["model"]["anchors"]) > 2:
             overlap_patches = config["model"]["anchors"][0]
 
+    #  Get normalization option
+    normalization_string = "STANDARD"
+    if "norm" in config["model"]:
+        normalization_string = config["model"]["norm"]
 
     write_direct = True
     if filament_mode:
@@ -470,6 +493,7 @@ def main(args=None):
     try:
         picking_results = do_prediction(
             config_path=config_path,
+            config_pre=config,
             weights_path=weights_path,
             input_path=input_path,
             obj_threshold=obj_threshold,
@@ -481,13 +505,15 @@ def main(args=None):
             num_gpus=num_gpus,
             num_cpus=num_cpus,
             otf=otf,
+            normalization=normalization_string,
             normalization_margin=norm_margin,
             write_direct=write_direct,
             min_distance=min_distance,
             outdir=outdir,
             monitor=monitor,
             min_size=min_size,
-            max_size=max_size
+            max_size=max_size,
+            skip_picked=args.skip,
         )
     except tf.errors.ResourceExhaustedError:
         print("############################")
@@ -497,8 +523,12 @@ def main(args=None):
 
     export_size = config_tools.get_box_size(config)[0]
     if write_direct == False:
-        del_not_fully_immersed, del_min_distance, del_min_size  = picking_postprocessing(
-            picking_results, export_size, min_distance, min_size=min_size,max_size=max_size
+        del_not_fully_immersed, del_min_distance, del_min_size = picking_postprocessing(
+            picking_results,
+            export_size,
+            min_distance,
+            min_size=min_size,
+            max_size=max_size,
         )
 
         if min_distance > 0:
@@ -520,7 +550,7 @@ def main(args=None):
             "particles as they were not out of specified size range",
         )
 
-    write_size_distribution_to_disk(picking_results,os.path.join(outdir,"DISTR"))
+    write_size_distribution_to_disk(picking_results, os.path.join(outdir, "DISTR"))
     ###############################
     #   Filament Post Processing
     ###############################
@@ -621,23 +651,11 @@ def main(args=None):
                             nosplit=nosplit,
                         )
                     )
-                """
-                procs = []
-                for tracer in filament_tracers:
-                    proc = multiprocessing.Process(target=tracer.trace_filaments)
-                    procs.append(proc)
-                    proc.start()
-                for proc in procs:
-                    proc.join()
-                """
-                #subset_new_filaments = [
-                #    tracer.trace_filaments() for tracer in filament_tracers
-                #]
 
                 pool = multiprocessing.Pool(processes=num_cpus)
 
                 subset_new_filaments = pool.map(
-                   trace_subset_filements, range(len(filament_tracers))
+                    trace_subset_filements, range(len(filament_tracers))
                 )
                 pool.close()
                 pool.join()
@@ -721,8 +739,17 @@ def main(args=None):
         if write_direct == False:
             prediction_result_to_disk(outdir, picking_results)
 
+    if args.cleanup:
+        print("#####################################")
+        print("Delete filtered images...")
+        if os.path.exists(config["model"]["filter"][-1]):
+            import shutil
+            shutil.rmtree(config["model"]["filter"][-1])
+        print("Done")
+        print("#####################################")
 
-def write_size_distribution_to_disk(picking_results, output_folder = ""):
+
+def write_size_distribution_to_disk(picking_results, output_folder=""):
 
     if not os.path.exists(output_folder):
         os.makedirs(output_folder)
@@ -733,7 +760,7 @@ def write_size_distribution_to_disk(picking_results, output_folder = ""):
         for box in boxes:
             est_width = box.meta["boxsize_estimated"][0]
             est_height = box.meta["boxsize_estimated"][1]
-            estimated_size.append((est_height+est_width)//2)
+            estimated_size.append((est_height + est_width) // 2)
             confidence.append(box.c)
     #################################
     # Size distribution histogram
@@ -746,32 +773,36 @@ def write_size_distribution_to_disk(picking_results, output_folder = ""):
     print("#####################################")
     print("")
     print("## Particle diameter distribution ##")
-    print("MEAN:", est_size_mean , "px")
-    print("SD:", est_size_sd ,"px")
-    print("25%-Quantile:", est_size_25q,"px")
-    print("50%-Quantile:", est_size_50q,"px")
-    print("75%-Quantile:", est_size_75q,"px")
+    print("MEAN:", est_size_mean, "px")
+    print("SD:", est_size_sd, "px")
+    print("25%-Quantile:", est_size_25q, "px")
+    print("50%-Quantile:", est_size_50q, "px")
+    print("75%-Quantile:", est_size_75q, "px")
 
-
-
-    import matplotlib.pyplot as pl
     import matplotlib as mpl
+
+    mpl.use("Agg")
+    import matplotlib.pyplot as pl
+
     mpl.rcParams["figure.dpi"] = 200
     mpl.rcParams.update({"font.size": 7})
-    width = max(10, int((np.max(estimated_size)-np.min(estimated_size))/10))
+    width = max(10, int((np.max(estimated_size) - np.min(estimated_size)) / 10))
     pl.hist(estimated_size, bins=width)
     pl.title("Particle diameter distribution")
-    pl.xlabel("Partilce diameter [px] (Bin size: "+str(width)+"px )")
+    pl.xlabel("Partilce diameter [px] (Bin size: " + str(width) + "px )")
     pl.ylabel("Count")
 
-    import time
     timestr = time.strftime("%Y%m%d-%H%M%S")
     import csv
 
-    output_path_distr_raw_txt = os.path.join(output_folder, "size_distribution_raw_" + timestr + ".txt")
-    np.savetxt(output_path_distr_raw_txt,estimated_size,fmt="%d")
-    output_path_distr_txt = os.path.join(output_folder, "size_distribution_summary_" + timestr + ".txt")
-    with open(output_path_distr_txt, 'w', newline='') as file:
+    output_path_distr_raw_txt = os.path.join(
+        output_folder, "size_distribution_raw_" + timestr + ".txt"
+    )
+    np.savetxt(output_path_distr_raw_txt, estimated_size, fmt="%d")
+    output_path_distr_txt = os.path.join(
+        output_folder, "size_distribution_summary_" + timestr + ".txt"
+    )
+    with open(output_path_distr_txt, "w", newline="") as file:
         writer = csv.writer(file)
         writer.writerow(["Metric", "Value"])
         writer.writerow(["MEAN", est_size_mean])
@@ -780,23 +811,28 @@ def write_size_distribution_to_disk(picking_results, output_folder = ""):
         writer.writerow(["Q50", est_size_50q])
         writer.writerow(["Q75", est_size_75q])
     print("Wrote particle size distribution summary to", output_path_distr_txt)
-    output_path_distr_img = os.path.join(output_folder,"size_distribution_"+timestr+".png")
+    output_path_distr_img = os.path.join(
+        output_folder, "size_distribution_" + timestr + ".png"
+    )
     print("Wrote plot of particle size distribution to", output_path_distr_img)
     pl.savefig(output_path_distr_img)
     pl.close()
 
-
     #################################
     # Confidence distribution histogram
     #################################
-    output_path_conf_img = os.path.join(output_folder, "confidence_distribution_" + timestr + ".png")
-    output_path_conf_raw = os.path.join(output_folder,
-                                        "confidence_distribution_raw_" + timestr + ".txt")
-    output_path_conf_sum = os.path.join(output_folder,
-                                        "confidence_distribution_summary_" + timestr + ".txt")
+    output_path_conf_img = os.path.join(
+        output_folder, "confidence_distribution_" + timestr + ".png"
+    )
+    output_path_conf_raw = os.path.join(
+        output_folder, "confidence_distribution_raw_" + timestr + ".txt"
+    )
+    output_path_conf_sum = os.path.join(
+        output_folder, "confidence_distribution_summary_" + timestr + ".txt"
+    )
 
     # RAW DATA
-    np.savetxt(output_path_conf_raw,confidence,fmt='%1.2f')
+    np.savetxt(output_path_conf_raw, confidence, fmt="%1.2f")
 
     # SUMMARY
     conf_mean = np.mean(confidence)
@@ -805,7 +841,7 @@ def write_size_distribution_to_disk(picking_results, output_folder = ""):
     conf_50q = np.percentile(confidence, 50)
     conf_75q = np.percentile(confidence, 75)
 
-    with open(output_path_conf_sum, 'w', newline='') as file:
+    with open(output_path_conf_sum, "w", newline="") as file:
         writer = csv.writer(file)
         writer.writerow(["Metric", "Value"])
         writer.writerow(["MEAN", conf_mean])
@@ -818,7 +854,7 @@ def write_size_distribution_to_disk(picking_results, output_folder = ""):
     width = max(10, int((np.max(confidence) - np.min(confidence)) / 0.05))
     pl.hist(confidence, bins=width)
     pl.title("Confidence distribution")
-    bin_size_str = "{0:.2f}".format(((np.max(confidence) - np.min(confidence))/width))
+    bin_size_str = "{0:.2f}".format(((np.max(confidence) - np.min(confidence)) / width))
     pl.xlabel("Confidence (Bin size: " + bin_size_str + ")")
     pl.ylabel("Count")
     pl.savefig(output_path_conf_img)
@@ -830,6 +866,7 @@ def write_size_distribution_to_disk(picking_results, output_folder = ""):
     print("#####################################")
     pl.close()
 
+
 def prediction_result_to_disk(outdir, picking_results):
     for box_to_write in picking_results:
 
@@ -876,7 +913,9 @@ def min_distance_filter(boxes, min_distance):
     all_comb = list(itertools.combinations(boxes, 2))
 
     distsqs = list(itertools.starmap(utils.box_squared_distance, all_comb))
-    low_distance_pairs = list(itertools.compress(all_comb, [distsq < min_distance_sq for distsq in distsqs]))
+    low_distance_pairs = list(
+        itertools.compress(all_comb, [distsq < min_distance_sq for distsq in distsqs])
+    )
     for box_a, box_b in low_distance_pairs:
         box_to_delte = box_a
         if box_a.c > box_b.c:
@@ -888,8 +927,10 @@ def min_distance_filter(boxes, min_distance):
 
 
 def rescale(box, image_height, image_width, export_size=None):
-    x_ll = int(box.x * image_width - box.w/2 * image_height) # lower left
-    y_ll = int(image_height - box.y * image_height - box.h / 2.0 * image_width) # lower right
+    x_ll = int(box.x * image_width - box.w / 2 * image_height)  # lower left
+    y_ll = int(
+        image_height - box.y * image_height - box.h / 2.0 * image_width
+    )  # lower right
     boxheight_in_pxl = int(box.h * image_width)
     boxwidth_in_pxl = int(box.w * image_height)
     if export_size is not None:
@@ -908,7 +949,9 @@ def rescale(box, image_height, image_width, export_size=None):
     return box
 
 
-def picking_postprocessing(picking_results, export_size, min_distance, min_size=None, max_size=None):
+def picking_postprocessing(
+    picking_results, export_size, min_distance, min_size=None, max_size=None
+):
 
     # Rescaling
     for picking_result_micrograph in picking_results:
@@ -920,11 +963,15 @@ def picking_postprocessing(picking_results, export_size, min_distance, min_size=
 
         for box in picking_result_micrograph["boxes"]:
             box.meta["boxsize_estimated"] = (
-                int(box.w * image_width), int(box.h * image_height))
+                int(box.w * image_width),
+                int(box.h * image_height),
+            )
 
         for box in picking_result_micrograph["boxes_unfiltered"]:
             box.meta["boxsize_estimated"] = (
-                int(box.w * image_width), int(box.h * image_height))
+                int(box.w * image_width),
+                int(box.h * image_height),
+            )
 
         # Resize box
         picking_result_micrograph["boxes"] = [
@@ -980,7 +1027,9 @@ def picking_postprocessing(picking_results, export_size, min_distance, min_size=
     for picking_result_micrograph in picking_results:
 
         boxes = picking_result_micrograph["boxes"]
-        out_of_size_box_indices = get_out_of_sizes_boxes(boxes,min_size=min_size,max_size=max_size)
+        out_of_size_box_indices = get_out_of_sizes_boxes(
+            boxes, min_size=min_size, max_size=max_size
+        )
 
         for index in sorted(out_of_size_box_indices, reverse=True):
             del boxes[index]
@@ -1004,8 +1053,10 @@ def do_prediction(
     num_cpus=-1,
     yolo=None,
     otf=False,
+    normalization="STANDARD",
     normalization_margin=0,
     monitor=False,
+    skip_picked=False,
     **kwargs
 ):
     """
@@ -1032,6 +1083,8 @@ def do_prediction(
         if not path_exists:
             sys.exit("Input path does not exist: " + path)
 
+
+
     img_paths = []
     if isinstance(input_path, list):
 
@@ -1068,6 +1121,12 @@ def do_prediction(
                 and os.path.isfile(os.path.join(input_path, i))
                 and i.endswith(("tiff", "tif", "mrc", "mrcs", "png", "jpg", "jpeg"))
             ]
+
+    if skip_picked:
+        import glob
+        already_picked = glob.glob(os.path.join(kwargs["outdir"], "*/*.box"))
+        already_picked = [os.path.splitext(os.path.basename(box_path))[0] for box_path in already_picked]
+        img_paths = [img_path for img_path in img_paths if os.path.splitext(os.path.basename(img_path))[0] not in already_picked]
     import time
 
     if monitor:
@@ -1116,15 +1175,38 @@ def do_prediction(
         # When monitoring a directory, wait for the first valid image
 
     img_paths.sort()
-    first_image_path = img_paths[0]
+
+    from cryolo.preprocessing import get_image_size_distr
+
+    size_distr = get_image_size_distr(img_paths)
+    if len(size_distr)>1:
+        print("Prediction on mixed-size images is not supported in single prediction run.")
+        print("The following training image sizes were detected:")
+        for size in size_distr:
+            print(size[0], "x", size[1], "( N:", size[2], ")")
+        print("Stop crYOLO.")
+        sys.exit(1)
+
+
     if config_pre is not None:
-        config = config_pre
+        config = config_pre.copy()
     else:
         with open(config_path) as config_buffer:
             config = json.load(config_buffer)
 
-    # Read (first) image and check the image depth.
+    cryolo_mode = utils.get_cryolo_mode(size_distr=size_distr, config_input_size=config["model"]["input_size"])
+    if cryolo_mode == "NON-SQUARE":
+        ar = size_distr[0][0]/size_distr[0][1]
+        if ar < 1:
+            input_size = [int(config["model"]["input_size"] / ar), config["model"]["input_size"]]
+        else:
+            input_size = [config["model"]["input_size"], int(config["model"]["input_size"] * ar)]
+        adjusted_size = config_tools.adjust_size(input_size)
+        print("Your training input size",config["model"]["input_size"],"was adjusted to",adjusted_size)
+        config["model"]["input_size"] = adjusted_size
 
+    # Read (first) image and check the image depth.
+    first_image_path = img_paths[0]
     try:
         img_first = imagereader.image_read(first_image_path)
     except ValueError:
@@ -1143,27 +1225,28 @@ def do_prediction(
     else:
         depth = 3
 
-    grid_w, grid_h = config_tools.get_gridcell_dimensions(config)
+    #grid_w, grid_h = config_tools.get_gridcell_dimensions(config)
 
     #############################################
     # Read meta data about the model
     #############################################
     anchors = None
     yolo_kwargs = {}
+    cryolo_version = None
     import h5py
 
     with h5py.File(weights_path, mode="r") as f:
-        try:
+        if "anchors" in f:
             anchors = list(f["anchors"])
-        except KeyError:
-            None
-
-    with h5py.File(weights_path, mode="r") as f:
-        try:
+        if "num_free_layers" in f:
             num_free_layers = int(list(f["num_free_layers"])[0])
             yolo_kwargs["num_fine_tune_layers"] = num_free_layers
-        except KeyError:
-            pass
+        if "cryolo_version" in f:
+            cryolo_version = list(f["cryolo_version"])[0].decode("utf-8")
+
+
+    if cryolo_version is not None:
+        print("Load crYOLO model which was trained with", cryolo_version)
 
     if anchors is None:
         #
@@ -1216,7 +1299,7 @@ def do_prediction(
     if num_patches == 1:
         # In case only one patch is used (should be default), the resizing can already
         # be done at the filtering step
-        resize_to = (config["model"]["input_size"], config["model"]["input_size"])
+        resize_to = config["model"]["input_size"]
 
     if monitor:
         # Monitor mode only works OTF
@@ -1251,6 +1334,7 @@ def do_prediction(
             # Normal lowpass filter
             cutoff, filter_img_path = filter_options
             if not otf:
+                start_f = time.time()
                 img_paths_filtered = filter_images_lowpass(
                     img_paths=img_paths,
                     output_dir_filtered_imgs=filter_img_path,
@@ -1258,6 +1342,8 @@ def do_prediction(
                     num_cpus=num_cpus,
                     resize_to=resize_to,
                 )
+                end_f = time.time()
+                print("Time needed for filtering:", end_f - start_f)
             else:
                 img_paths_filtered = img_paths
     else:
@@ -1304,7 +1390,6 @@ def do_prediction(
                             )
 
                             if resize_to is not None:
-                                print("Resize!")
                                 from PIL import Image
 
                                 image = np.array(
@@ -1378,6 +1463,7 @@ def do_prediction(
                             nms_threshold=nms_thresh,
                             num_patches=num_patches,
                             normalize_margin=normalization_margin,
+                            normalization=normalization
                         )
                         end = time.time()
                         measured_times.append(end - start)
@@ -1439,12 +1525,16 @@ def do_prediction(
                                         export_size = config_tools.get_box_size(config)[
                                             0
                                         ]
-                                        del_not_fully_immersed, del_min_distance, del_min_size = picking_postprocessing(
+                                        (
+                                            del_not_fully_immersed,
+                                            del_min_distance,
+                                            del_min_size,
+                                        ) = picking_postprocessing(
                                             picking_results=[box_to_write],
                                             export_size=export_size,
                                             min_distance=kwargs["min_distance"],
-                                            min_size= kwargs["min_size"],
-                                            max_size=kwargs["max_size"]
+                                            min_size=kwargs["min_size"],
+                                            max_size=kwargs["max_size"],
                                         )
                                         sum_del_min_distance += del_min_distance
                                         sum_del_not_fully_immersed += (
@@ -1487,9 +1577,11 @@ def do_prediction(
         )
         for img in skipped_images:
             print(img)
+
+    print("#####################################")
     print(
         total_picked,
-        "particles in total are found",
+        "particles in total has been found",
         "(",
         int(np.sum(measured_times)),
         "seconds)",
@@ -1512,13 +1604,13 @@ def do_prediction(
         print(
             "Deleted",
             sum_del_out_of_size,
-            "particles as they were not out of specified size range",
+            "particles as they were out of specified size range",
         )
-
+    print("#####################################")
     return boxes_to_write
 
 
-def get_out_of_sizes_boxes(boxes,min_size=None,max_size=None):
+def get_out_of_sizes_boxes(boxes, min_size=None, max_size=None):
     if min_size is None and max_size is None:
         return []
 
@@ -1526,7 +1618,7 @@ def get_out_of_sizes_boxes(boxes,min_size=None,max_size=None):
     for box_index, box in enumerate(boxes):
         box_width = box.meta["boxsize_estimated"][0]
         box_height = box.meta["boxsize_estimated"][1]
-        size = (box_width+box_height)/2
+        size = (box_width + box_height) / 2
         out_of_size = False
 
         if min_size:
@@ -1592,14 +1684,15 @@ def filter_images_lowpass(
                     )
                 )
 
-    num_processes = None
+    num_processes = int(multiprocessing.cpu_count() / 2)
     if num_cpus != -1:
         num_processes = num_cpus
-    lock = multiprocessing.Lock()
+
+    print("oO", len(arg_tubles),num_processes)
+    maxtaskperchild = max(1,len(arg_tubles)//num_processes)
+
     pool = multiprocessing.Pool(
-        initializer=lowpass.init,
-        initargs=(lock,),
-        maxtasksperchild=1,
+        maxtasksperchild=maxtaskperchild,
         processes=num_processes,
     )
 
