@@ -28,14 +28,13 @@ import argparse
 import json
 import multiprocessing
 import os
-import time
 
 import numpy as np
 
 import cryolo.config_tools as config_tools
 import cryolo.imagereader as imagereader
 import cryolo.lowpass as lowpass
-from cryolo.preprocessing import parse_annotation2
+import cryolo.preprocessing as preprocess
 from . import utils
 from gooey import Gooey, GooeyParser
 
@@ -129,9 +128,9 @@ def create_parser(parser):
         action="store_true",
         default=False,
         help="Set it to true if you only want to use the fine tune mode. "
-             "When using the fine tune mode, only the last layers of your network are trained and you"
-             "have to specify pretrained_weights (see action \"config\"->\"Training options\") "
-             "You typically use a general model as pretrained weights."
+        "When using the fine tune mode, only the last layers of your network are trained and you"
+        'have to specify pretrained_weights (see action "config"->"Training options") '
+        "You typically use a general model as pretrained weights.",
     )
 
     optional_group.add_argument(
@@ -142,6 +141,13 @@ def create_parser(parser):
         help="Layers to be trained when using fine tuning.",
     )
 
+    optional_group.add_argument(
+        "--cleanup",
+        action="store_true",
+        default=False,
+        help="If true, it will delete the filtered images after training is done."
+    )
+
     deexsp_group = parser.add_argument_group(
         "Deprecated/Experimental/Special ",
         "Contains either deprecated / experimental or very special options.",
@@ -167,37 +173,12 @@ def create_parser(parser):
         help="Use it if you want to deactivate data augmentation during training.",
     )
 
-
-# Reset Keras Session
-def reset_keras(gpu_fraction=1):
-    """
-    This functions resets keras memories after training. It is necessary for training in a
-    loop (like warmup and real training). Otherwise the program get memory problems.
-    :param gpu_fraction: Fraction of memory that should be allocaded.
-    :return: None
-    """
-    from keras.backend.tensorflow_backend import set_session
-    from keras.backend.tensorflow_backend import clear_session
-    from keras.backend.tensorflow_backend import get_session
-    import gc
-    import tensorflow
-
-    sess = get_session()
-    clear_session()
-    sess.close()
-    sess = get_session()
-
-    print(
-        gc.collect()
-    )  # if it's done something you should see a number being outputted
-
-    # use the same config as you used to create the session
-    config = tensorflow.ConfigProto()
-    config.gpu_options.allow_growth = True
-    config.gpu_options.per_process_gpu_memory_fraction = gpu_fraction
-    # config.gpu_options.visible_device_list = "0"
-    set_session(tensorflow.Session(config=config))
-
+    deexsp_group.add_argument(
+        "--use_multithreading",
+        action="store_true",
+        default=False,
+        help="If python multiprocessing leads to problems during training (e.g. freezing, dying workers) use multithreading instead of multiprocessing.",
+    )
 
 def get_parser():
     """
@@ -211,7 +192,6 @@ def get_parser():
     create_parser(parser)
     return parser
 
-
 def get_model(
     config,
     depth,
@@ -289,8 +269,17 @@ def _main_():
     Fake main function, initilazies gooey
     :return: None
     """
+    start_method="spawn"
     try:
-        multiprocessing.set_start_method("spawn")
+        os_start_method = os.environ["CRYOLO_MP_START"]
+        if os_start_method in ["spawn","fork"]:
+            start_method = os_start_method
+    except:
+        pass
+
+    try:
+
+        multiprocessing.set_start_method(start_method)
     except RuntimeError:
         print("Ignore set start method")
 
@@ -324,13 +313,18 @@ def main(args=None):
     except RuntimeError:
         pass
     import cryolo.utils as util
+
     util.check_for_updates()
+
     print("START METHOD:", multiprocessing.get_start_method())
 
     if args is None:
         parser = get_parser()
         args = parser.parse_args()
 
+    ###############################
+    #   Setup which GPU is used
+    ###############################
     if isinstance(args.gpu, list):
         if len(args.gpu) == 1:
             str_gpus = args.gpu[0].strip().split(" ")
@@ -344,6 +338,9 @@ def main(args=None):
             str_gpus = str(args.gpu)
             os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(str_gpus)
 
+    ###############################
+    #   Setup amount of GPU memory used
+    ###############################
     if args.gpu_fraction < 1.0 and args.gpu_fraction > 0.0:
         import tensorflow as tf
         from keras.backend.tensorflow_backend import set_session
@@ -363,10 +360,8 @@ def main(args=None):
 
         set_session(tf.Session(config=config))
 
-    config_path = args.conf
-
     early_stop = int(args.early)
-
+    do_multiprocessing = args.use_multithreading == False
     warm_restarts = args.warm_restarts
 
     do_augmentation = not args.skip_augmentation
@@ -374,13 +369,63 @@ def main(args=None):
     do_fine_tune = args.fine_tune
     num_fine_tune_layers = args.layers_fine_tune
 
-    num_cpus = int(multiprocessing.cpu_count()/2)
+    num_cpus = int(multiprocessing.cpu_count() / 2)
     if args.num_cpu != -1:
         num_cpus = args.num_cpu
     seed = args.seed
+
+    experimental_loss = False
+
+
+
+    ###############################
+    #   Read configuration
+    ###############################
+    config_path = args.conf
+    with open(config_path) as config_buffer:
+        try:
+            config = json.loads(config_buffer.read())
+        except json.JSONDecodeError:
+            print(
+                "Your configuration file seems to be corruped. Please check if it is valid."
+            )
+
+    # Setup log dir
+    if "other" in config and "log_path" in config["other"]:
+        log_path = config["other"]["log_path"]
+    else:
+        log_path = "logs_cryolo/"
+        print("")
+    if not os.path.exists(log_path):
+        try:
+            os.makedirs(log_path)
+        except Exception:
+            import sys
+
+            print(
+                "Log path can't be created. Check your configuration ("
+                + config_path
+                + ")."
+            )
+            sys.exit(1)
+
+    ###############################
+    #   Write crYOLO command to disk
+    ###############################
+    import time
+    import sys
+
+    timestr = time.strftime("%Y%m%d-%H%M%S")
+    utils.write_command(
+        os.path.join(log_path, "cmdlogs/", "command_train_" + timestr + ".txt"),
+        "cryolo_train.py " + " ".join(sys.argv[1:]),
+    )
+
+    ###############################
+    #   Read number of warmup epochs
+    ###############################
     if args.warmup is not None:
         warmup_epochs = int(args.warmup)
-        print("Read warmup by argument")
     else:
         if config["train"]["warmup_epochs"] is not None:
             warmup_epochs = config["train"]["warmup_epochs"]
@@ -393,60 +438,83 @@ def main(args=None):
         warmup_epochs = 0
         print("Ignore warmup, as fine tuning is activated.")
 
+    # Early stop should be at least the number of warmup epochs
     if early_stop < warmup_epochs:
         early_stop = warmup_epochs
 
-    experimental_loss = False
-
-    #
-    # Read and adapt configuration
-    #
-
-    with open(config_path) as config_buffer:
-        try:
-            config = json.loads(config_buffer.read())
-        except json.JSONDecodeError:
-            print(
-                "Your configuration file seems to be corruped. Please check if it is valid."
-            )
-        if config["model"]["input_size"] % 32 > 0:
-            new_input_size = int(32 * round(float(config["model"]["input_size"]) / 32))
-            print(
-                "You input size is not a multiple of 32. Round it to the next multiple of 32:",
-                new_input_size,
-            )
-            config["model"]["input_size"] = new_input_size
     num_patches = config_tools.get_number_patches(config)
     experimental_loss = False
 
     if early_stop < warmup_epochs:
         early_stop = warmup_epochs
 
-    grid_w, grid_h = config_tools.get_gridcell_dimensions(config)
-
     ###############################
     #   Parse the annotations
     ###############################
 
+    # Identify all image - annotation pair
+    train_annot_pairs = preprocess.find_image_annotation_pairs_by_dir(os.path.realpath(config["train"]["train_annot_folder"]),
+                                                                      os.path.realpath(config["train"]["train_image_folder"]))
+
+    # Identifiy all image sizes:
+    sizes = preprocess.get_image_size_distr([pair[0] for pair in train_annot_pairs])
+
+    cryolo_mode = utils.get_cryolo_mode(size_distr=sizes, config_input_size=config["model"]["input_size"])
+
+    print("###############################################")
+    print("The following training image sizes were detected:")
+    for size in sizes:
+        print(size[0], "x", size[1], "( N:",size[2],")")
+    print("")
+    print("crYOLO will train in mode:", cryolo_mode)
+
+
+
+
+    if cryolo_mode == "NON-SQUARE":
+        width = sizes[0][0]
+        height = sizes[0][1]
+        ar = height/width
+        if ar < 1:
+            input_size = [config["model"]["input_size"],int(config["model"]["input_size"] / ar)]
+        else:
+            input_size = [int(config["model"]["input_size"] * ar),config["model"]["input_size"]]
+
+        adjusted_size = config_tools.adjust_size(input_size)
+        print("Your training input size",config["model"]["input_size"],"was adjusted to",adjusted_size,"(height, width)")
+        config["model"]["input_size"]= adjusted_size
+    else:
+        config["model"]["input_size"] = config_tools.get_adjusted_input_size(config)
+
+    print("###############################################")
+
+    grid_w, grid_h = config_tools.get_gridcell_dimensions(config)
+
     # parse annotations of the training set
-    train_imgs, train_labels = parse_annotation2(
-        ann_dir=os.path.realpath(config["train"]["train_annot_folder"]),
-        img_dir=os.path.realpath(config["train"]["train_image_folder"]),
+    train_imgs, train_labels = preprocess.parse_annotation(
+        train_annot_pairs,
         grid_dims=(grid_w, grid_h, num_patches),
         anchor_size=int(config["model"]["anchors"][0]),
     )
+
     config["model"]["labels"] = ["particle"]
 
     if len(train_imgs) == 0:
         print("No image files were found.")
         import sys
+
         sys.exit(0)
     # parse annotations of the validation set, if any, otherwise split the training set
     if os.path.exists(config["valid"]["valid_annot_folder"]):
-        valid_imgs, valid_labels = parse_annotation2(
+
+        valid_annot_pairs = preprocess.find_image_annotation_pairs_by_dir(
             os.path.realpath(config["valid"]["valid_annot_folder"]),
-            os.path.realpath(config["valid"]["valid_image_folder"]),
-            (grid_w, grid_h, num_patches),
+            os.path.realpath(config["valid"]["valid_image_folder"]))
+
+        valid_imgs, valid_labels = preprocess.parse_annotation(
+            valid_annot_pairs,
+            grid_dims=(grid_w, grid_h, num_patches),
+            anchor_size=int(config["model"]["anchors"][0]),
         )
 
         if (
@@ -468,8 +536,7 @@ def main(args=None):
         np.random.shuffle(train_imgs)
         valid_imgs = train_imgs[train_valid_split:]
         train_imgs = train_imgs[:train_valid_split]
-        print("Validation set:")
-        print([item["filename"] for item in valid_imgs])
+        print("Validation set:", len(valid_imgs),"images")
 
     #####################################
     # Write runfile
@@ -487,11 +554,14 @@ def main(args=None):
     runjson["run"]["train_images"] = train_imgs_paths
     runjson["run"]["train_annot"] = train_annot_paths
 
-    if not os.path.exists("runfiles/"):
-        os.mkdir("runfiles/")
+    runfiles_path = os.path.join(log_path, "runfiles/")
+    if not os.path.exists(runfiles_path):
+        os.mkdir(runfiles_path)
     timestr = time.strftime("%Y%m%d-%H%M%S")
-    with open("runfiles/run_" + timestr + ".json", "w") as outfile:
+    runfile_path = os.path.join(runfiles_path, timestr + ".json")
+    with open(runfile_path, "w") as outfile:
         json.dump(runjson, outfile, ensure_ascii=False, indent=4)
+    print("Wrote runfile to: " + runfile_path)
 
     ##############################
     # Filter the data
@@ -501,7 +571,7 @@ def main(args=None):
     if num_patches == 1:
         # In case only one patch is used (should be default), the resizing can already
         # be done at the filtering step
-        resize_to = (config["model"]["input_size"], config["model"]["input_size"])
+        resize_to = config["model"]["input_size"]
 
     if "filter" in config["model"]:
         filter_options = config["model"]["filter"]
@@ -546,22 +616,18 @@ def main(args=None):
                 filter_cutoff=cutoff,
                 filter_tmp_path=filter_img_path,
                 num_cpu=num_cpus,
-                resize_to=resize_to,
+                resize_to_shape=resize_to,
             )
             lowpass.filter_images(
                 images=valid_imgs,
                 filter_cutoff=cutoff,
                 filter_tmp_path=filter_img_path,
                 num_cpu=num_cpus,
-                resize_to=resize_to,
+                resize_to_shape=resize_to,
             )
             end = time.time()
             print("Time needed for filtering:", end - start)
 
-    overlap_labels = set(config["model"]["labels"]).intersection(
-        set(train_labels.keys())
-    )
-
     # Read first image and check the image depth.
     img_first = imagereader.image_read(train_imgs[0]["filename"])
     isgrey = False
@@ -578,36 +644,7 @@ def main(args=None):
         depth = 3
 
     # As only on box size is expected, the anchor box size automatically
-
-    if len(config["model"]["anchors"]) > 2:
-        # general model is used
-        img_width = 4096.0 / num_patches
-        img_height = 4096.0 / num_patches
-        cell_w = img_width / grid_w
-        cell_h = img_height / grid_h
-        anchors = np.array(config["model"]["anchors"], dtype=float)
-        anchors[::2] = anchors[::2] / cell_w
-        anchors[1::2] = anchors[1::2] / cell_h
-
-    else:
-        # specifc model is used
-        img_width = float(train_imgs[0]["img_size"][0]) / num_patches
-        img_height = float(train_imgs[0]["img_size"][1]) / num_patches
-
-        cell_w = img_width / grid_w
-        cell_h = img_height / grid_h
-        box_width, box_height = config_tools.get_box_size(config)
-
-        anchor_width = 1.0 * box_width / cell_w
-        anchor_height = 1.0 * box_height / cell_h
-        anchors = [anchor_width, anchor_height]
-
-    if not config["train"]["log_path"]:
-        log_path = "~/logs/"
-    else:
-        log_path = config["train"]["log_path"]
-        if not os.path.exists(log_path):
-            os.mkdir(log_path)
+    anchors = utils.get_anchors(config, image_sizes=sizes)#
 
     # Get overlap patches
     overlap_patches = 0
@@ -616,6 +653,12 @@ def main(args=None):
     elif not len(config["model"]["anchors"]) > 2:
         overlap_patches = config["model"]["anchors"][0]
 
+
+    #  Get normalization option
+    normalization_string = "STANDARD"
+    if "norm" in config["model"]:
+        normalization_string = config["model"]["norm"]
+
     ###############################
     # Start the training process
     ###############################
@@ -623,7 +666,6 @@ def main(args=None):
     if os.path.exists(config["train"]["pretrained_weights"]):
         pretrained_weights = config["train"]["pretrained_weights"]
 
-    from keras import backend as K
     import tensorflow as tf
 
     start = time.time()
@@ -646,6 +688,9 @@ def main(args=None):
 
         parallel_model = multi_gpu_model(yolo.model, gpus=num_gpus, cpu_merge=False)
     try:
+        tf_log_path = os.path.join(log_path, "tensorflow/")
+        if not os.path.exists(tf_log_path):
+            os.mkdir(tf_log_path)
         yolo.train(
             train_imgs=train_imgs,
             valid_imgs=valid_imgs,
@@ -661,7 +706,7 @@ def main(args=None):
             class_scale=config["train"]["class_scale"],
             saved_weights_name=config["train"]["saved_weights_name"],
             debug=config["train"]["debug"],
-            log_path=log_path,
+            log_path=tf_log_path,
             early_stop_thresh=early_stop,
             num_patches=num_patches,
             warm_restarts=warm_restarts,
@@ -669,20 +714,39 @@ def main(args=None):
             parallel_model=parallel_model,
             num_cpus=num_cpus,
             do_augmentation=do_augmentation,
+            do_multiprocessing=do_multiprocessing,
+            normalization=normalization_string
         )
     except tf.errors.ResourceExhaustedError:
         print("############################")
         print("Not enough GPU memory. Try to reduce training batch size. ")
         print("############################")
         import sys
+
         sys.exit(0)
 
     end = time.time()
 
     print("Time elapsed for training:", (end - start))
 
+    if args.cleanup:
+        print("#########################")
+        print("Delete filtered images...")
+        if os.path.exists(config["model"]["filter"][-1]):
+            import shutil
+            shutil.rmtree(config["model"]["filter"][-1])
+        print("Done")
+        print("#########################")
 
 
 if __name__ == "__main__":
-    multiprocessing.set_start_method("spawn")
+    start_method = "spawn"
+    try:
+        os_start_method = os.environ["CRYOLO_MP_START"]
+
+        if os_start_method in ["spawn","fork"]:
+            start_method = os_start_method
+    except:
+        pass
+    multiprocessing.set_start_method(start_method)
     _main_()
